/*
 * Project: irond
 * Package: src.de.fhhannover.inform.ifmap.transform
 * File:    MessageTransformer.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.ifmap.transform;
/**
 * Transform JAXB-RI Objects to JAVA-Objects
 */
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.UUID;

import org.apache.xerces.dom.ElementNSImpl;
import org.trustedcomputinggroup._2010.ifmap._2.AccessRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.DeleteResultType;
import org.trustedcomputinggroup._2010.ifmap._2.DeviceType;
import org.trustedcomputinggroup._2010.ifmap._2.ErrorResultType;
import org.trustedcomputinggroup._2010.ifmap._2.IPAddressType;
import org.trustedcomputinggroup._2010.ifmap._2.IdentityType;
import org.trustedcomputinggroup._2010.ifmap._2.MACAddressType;
import org.trustedcomputinggroup._2010.ifmap._2.MetadataListType;
import org.trustedcomputinggroup._2010.ifmap._2.ObjectFactory;
import org.trustedcomputinggroup._2010.ifmap._2.PollResultType;
import org.trustedcomputinggroup._2010.ifmap._2.ResultItemType;
import org.trustedcomputinggroup._2010.ifmap._2.SearchResultType;
import org.trustedcomputinggroup._2010.ifmap._2.SubscribeRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.UpdateResultType;
import org.trustedcomputinggroup._2010.ifmap._2.SubscribeRequestType.Update;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import de.fhhannover.inform.ifmap.communication.SOAPConnector;
import de.fhhannover.inform.ifmap.datastructure.Identifier;
import de.fhhannover.inform.ifmap.datastructure.IdentifierContainer;
import de.fhhannover.inform.ifmap.datastructure.Link;
import de.fhhannover.inform.ifmap.datastructure.LinkContainer;
import de.fhhannover.inform.ifmap.event.IdentifierChangedReceiver;
import de.fhhannover.inform.ifmap.event.StatusChangedReceiver;
import de.fhhannover.inform.ifmap.util.Crypto;

public class MessageTransformer {
	private ObjectFactory ifmap_factory = new ObjectFactory();
	private IdentifierContainer search_identifier = new IdentifierContainer();	
	private ArrayList<StatusChangedReceiver> receiver_status = new ArrayList<StatusChangedReceiver>();
	private ArrayList<IdentifierChangedReceiver> receiver_identifier = new ArrayList<IdentifierChangedReceiver>();
	private HashMap<String, Identifier> subscribtions = new HashMap<String, Identifier>();
	private Hashtable<String, Link> str2link = new Hashtable<String, Link>();
	private SOAPConnector soap;

	private Poller poller = new Poller();
	private boolean isDone = false;

	public MessageTransformer() {
		soap = new SOAPConnector(this);
	}

	public boolean connect() throws RemoteException {
		return soap.connect();
	}
	
	public void disconnect() {
		soap.disconnect();
	}
	
	public String getPublisherID(){
		return soap.getPublisherID();
	}
	
	public String getSessionID(){
		return soap.getSessionID();
	}
	
	public void poll() 
	throws RemoteException {		
		PollResultType result = soap.sendPoll();		
		if (result != null) {				
			List<Object> list = result.getSearchResultOrUpdateResultOrDeleteResult();			
			for (Object jaxb : list) {
				IdentifierContainer conti = new IdentifierContainer();	
				if(jaxb instanceof ErrorResultType){					
					continue;
				}							
				
				List<ResultItemType> rit;
				String rName;
				
				if(jaxb instanceof UpdateResultType){
					UpdateResultType srt = (UpdateResultType)jaxb;
					rit = srt.getResultItem();					
					rName = "updateResult";
				}
				else if(jaxb instanceof SearchResultType){
					SearchResultType srt = (SearchResultType) jaxb;
					rit = srt.getResultItem();
					rName = "searchResult";
				}
				else if(jaxb instanceof DeleteResultType){
					DeleteResultType srt = (DeleteResultType) jaxb;
					rit = srt.getResultItem();					
					rName = "deleteResult";
				}
				else return;
				
				if (result != null) {					
					processSearchResult(rit, conti);					
					notifyReceiver(rName, conti);				
				}				
			}			
		}
	}
	
	private void notifyReceiver(String name, IdentifierContainer conti){		
		if(name.equals("searchResult")){			
			this.notifyIdentifierNew(conti);
		}
		else if(name.equals("updateResult") || name.equals("notifyResult")){			
			this.notifyIdentifierChanged(conti);
		}
		else if(name.equals("deleteResult")){			
			this.notifyIdentifierDelete(conti);
		}
	}
	
	public void enablePolling(boolean poll) {		
		if (!poller.isAlive()) {			
			if (poll) {		
				poller = new Poller();
				poller.start();
			}
		} else {
			if (!poll) {
				isDone = true;
			}
		}

	}

	public boolean subscribeUpdate(Object identifier) throws RemoteException {
		SubscribeRequestType srt = ifmap_factory.createSubscribeRequestType();
		Update update = ifmap_factory.createSubscribeRequestTypeUpdate();
		
		String _name = UUID.randomUUID().toString().substring(0, 16);
		
		if(identifier instanceof IPAddressType) {
				IPAddressType ip = (IPAddressType) identifier;
				update.setIpAddress(ip);				
		}
		else if(identifier instanceof AccessRequestType) {
			AccessRequestType ar = (AccessRequestType) identifier;
			update.setAccessRequest(ar);				
		}
		else if(identifier instanceof MACAddressType) {
			MACAddressType mac = (MACAddressType) identifier;
			update.setMacAddress(mac);			
		}
		else if(identifier instanceof DeviceType) {
			DeviceType d = (DeviceType) identifier;
			update.setDevice(d);	
		}
		else if(identifier instanceof IdentityType) {
			IdentityType id = (IdentityType) identifier;
			update.setIdentity(id);	
		}
		else return false;
		
		update.setName(_name);
		update.setMaxDepth(10l);
		
		if(!subscribtions.containsKey(_name)){
			srt.getUpdateOrDelete().add(update);
			subscribtions.put(_name, null);
		}
		
		return soap.subscribeUpdate(srt);
		
	}
	
	public boolean subscribeUpdate(IdentifierContainer conti) throws RemoteException {
		SubscribeRequestType srt = ifmap_factory.createSubscribeRequestType();
		for (Identifier ident : conti) {
			Update update = this.ifmap_factory
					.createSubscribeRequestTypeUpdate();
			
			// we calculate the name for the subscription
			String _name = ident.getName()+ident.getAttributes();
			// create a md5 hash and take every second character for a string with length of 16
			String md5 = Crypto.generateMD5(_name);			
			StringBuffer buffer = new StringBuffer();
			for(int i = 0; i < md5.length(); i=i+2){
				buffer.append(md5.charAt(i));
			}
								
			update.setName(buffer.toString());
			update.setMaxDepth(10l);			
			Object o = ident.getRequestObject();
			if (o instanceof AccessRequestType) {
				update.setAccessRequest((AccessRequestType) o);
			} else if (o instanceof DeviceType) {
				update.setDevice((DeviceType) o);
			} else if (o instanceof IPAddressType) {
				IPAddressType ip = (IPAddressType) o;
				update.setIpAddress(ip);
			} else if (o instanceof IdentityType) {
				update.setIdentity((IdentityType) o);
			} else if (o instanceof MACAddressType) {
				update.setMacAddress((MACAddressType) o);
			} else {				
				return false;
			}
			
			if(!subscribtions.containsKey(_name)){
				srt.getUpdateOrDelete().add(update);
				subscribtions.put(_name, ident);
			}
		}
		return soap.subscribeUpdate(srt);
	}

	public IdentifierContainer getSearchIdentifier(String filter)
			throws RemoteException {
		search_identifier.clear();
		if (soap.connect()) {
			List<Object> list = soap.getDumpResult(filter);
			if (list == null) {				
				return null;
			} else {
				for (Object o : list) {
					Identifier id = new Identifier(o);
					search_identifier.add(id);
				}
			}
			return search_identifier;
		} else
			return null;
	}

	public long getLastUpdate() throws RemoteException {
		return soap.getLastUpdate();
	}

	public void addStatusListener(StatusChangedReceiver r) {
		if (!receiver_status.contains(r))
			receiver_status.add(r);
	}

	public void addUpdateListener(IdentifierChangedReceiver r) {
		if (!receiver_identifier.contains(r))
			receiver_identifier.add(r);
	}

	public synchronized void notifyIdentifierNew(IdentifierContainer conti) {
		for (IdentifierChangedReceiver r : receiver_identifier) {			
			r.processNew(conti);
		}
	}
	
	public synchronized void notifyIdentifierChanged(IdentifierContainer conti) {
		for (IdentifierChangedReceiver r : receiver_identifier) {
			r.processUpdate(conti);
		}
	}
	
	public synchronized void notifyIdentifierDelete(IdentifierContainer conti) {
		for (IdentifierChangedReceiver r : receiver_identifier) {
			r.processDelete(conti);
		}
	}

	public synchronized void notifyStatusChangedReceiver(String evt) {
		for (StatusChangedReceiver r : receiver_status) {
			r.processStatusEvent(evt);
		}
	}

	private void processSearchResult(List<ResultItemType> l, IdentifierContainer conti) {	
		if (l != null && !l.isEmpty()) {			
			for (ResultItemType result : l) {
				processResult(result, conti);
			}
		}		
	}

	private void processResult(ResultItemType result, IdentifierContainer conti) {
		MetadataListType meta = result.getMetadata();
		
		if(meta != null){
			if (result.getAccessRequestOrIdentityOrIpAddress().size() == 1) {
				Object o = result.getAccessRequestOrIdentityOrIpAddress().get(0);
				Identifier i = new Identifier(o);
				LinkContainer link = processMeta(i, i, meta);
				i.addLink(link);		
				conti.addIdentifier(i);		
			}

			if (result.getAccessRequestOrIdentityOrIpAddress().size() == 2) {				
				Object o1 = result.getAccessRequestOrIdentityOrIpAddress().get(0);
				Object o2 = result.getAccessRequestOrIdentityOrIpAddress().get(1);
				Identifier i1 = new Identifier(o1);
				Identifier i2 = new Identifier(o2);			
				LinkContainer container = processMeta(i1, i2, meta);
				i1.addLink(container);
				i2.addLink(container);
				conti.addIdentifier(i1);
				conti.addIdentifier(i2);			
			}
		}
	}

	private LinkContainer processMeta(Identifier i1, Identifier i2,
			MetadataListType metadata) {
		List<Object> ml = metadata.getAny();
		LinkContainer container = new LinkContainer();
		
		for (int i = 0; i < ml.size(); i++) {
			ElementNSImpl meta = (ElementNSImpl) ml.get(i);			
			Link link = new Link(meta.getNodeName(), i1, i2);		
			NamedNodeMap map = meta.getAttributes();
			for (int n = 0; n < map.getLength(); n++) {
				Node node = map.item(n);
				link.addAttribute(node.getNodeName(), node.getNodeValue());
			}
			
			NodeList nl = meta.getChildNodes();
			
			for (int j = 0; j < nl.getLength() - 1; j++) {
				Node n = nl.item(j);
				if (!n.hasAttributes()) {
					link.addMeta(n.getLocalName(), "value", n.getTextContent());
				} else {
					NamedNodeMap nnm = n.getAttributes();
					if (nnm != null) {
						for (int l = 0; l < nnm.getLength(); l++) {
							link.addMeta(n.getLocalName(), nnm.item(l)
									.getNodeName(), nnm.item(l).getNodeValue());							
						}
					}
				}
			}
			
			if(link.getAttribute("ifmap-cardinality").equals("singleValue")){
				link.setSingleValue(true);
			}
			else link.setSingleValue(false);
			
			if(!str2link.containsKey(link.toString())){				
				container.addLink(link);
				str2link.put(link.toString(), link);								
			}			
		}		
		return container;
	}

	private class Poller extends Thread {
		public Poller() {
		}

		@Override
		public void run() {
			while (!isDone) {
				//notifyStatusChangedReceiver("Starte Polling...");				
				try {					
					poll();
				} catch (RemoteException e) {
					notifyStatusChangedReceiver("Poll Error...");					
					try {
						sleep(10000);
					} catch (InterruptedException e1) {						
						e1.printStackTrace();
					}
				}				
			}			
		}
	}
}
